# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" global configuration settings """
import enum
import logging
import os
from typing import Literal, Union

from qiboconnection import __version__  # pylint: disable=cyclic-import

logger = logging.getLogger(__name__)


if "QUANTUM_SERVICE_URL" not in os.environ:
    QUANTUM_SERVICE_URL = {
        "local": "http://localhost:8080",
        "docker_local": "http://nginx:8080",
        "staging": "https://qilimanjaroqaas.ddns.net:8080",
        "development": "https://qilimanjarodev.ddns.net:8080",
    }


class EnvironmentType(str, enum.Enum):
    """Environment Type

    Args:
        enum (str): Available environment types:
            * local
            * staging
    """

    LOCAL = "local"
    STAGING = "staging"
    DEVELOPMENT = "development"
    LAMBDA = "lambda"


class Environment:
    """Execution Environment"""

    def __init__(self, environment_type: Union[EnvironmentType | None] = None):
        if "QUANTUM_SERVICE_URL" in os.environ:
            self._environment_type = EnvironmentType.LAMBDA
            self.quantum_service_url = os.environ["QUANTUM_SERVICE_URL"]
            self._audience_url = os.environ.get("AUDIENCE_URL", self.quantum_service_url)
        else:
            if environment_type not in [
                EnvironmentType.LOCAL,
                EnvironmentType.STAGING,
                EnvironmentType.DEVELOPMENT,
            ]:
                raise ValueError("Environment Type MUST be 'local', 'staging' or 'development'")
            if environment_type == EnvironmentType.LOCAL:
                self._environment_type = EnvironmentType.LOCAL
                self.quantum_service_url = QUANTUM_SERVICE_URL["local"]
                self._audience_url = QUANTUM_SERVICE_URL["docker_local"]
            if environment_type == EnvironmentType.STAGING:
                self._environment_type = EnvironmentType.STAGING
                self.quantum_service_url = QUANTUM_SERVICE_URL["staging"]
                self._audience_url = QUANTUM_SERVICE_URL["staging"]
            if environment_type == EnvironmentType.DEVELOPMENT:
                self._environment_type = EnvironmentType.DEVELOPMENT
                self.quantum_service_url = QUANTUM_SERVICE_URL["development"]
                self._audience_url = QUANTUM_SERVICE_URL["development"]

    @property
    def qibo_quantum_service_url(self) -> str:
        """returns quantum service url

        Returns:
            str: quantum service url
        """
        return self.quantum_service_url

    @property
    def audience_url(self) -> str:
        """returns the audience url

        Returns:
            str: audience url
        """
        return self._audience_url

    @property
    def environment_type(
        self,
    ) -> Union[
        Literal[EnvironmentType.LOCAL],
        Literal[EnvironmentType.STAGING],
        Literal[EnvironmentType.DEVELOPMENT],
        Literal[EnvironmentType.LAMBDA],
    ]:
        """Returns the environment_type

        Returns:
            Union[Literal[EnvironmentType.LOCAL], Literal[EnvironmentType.STAGING]]: environment type
        """
        return self._environment_type


def get_environment() -> Environment:
    """Return environment corresponding to QIBOCONNECTION_ENVIRONMENT env variable

    Returns:
        Environment: environment"""
    return Environment(environment_type=EnvironmentType(os.environ.get("QIBOCONNECTION_ENVIRONMENT", "staging")))



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Quantum Device """
from typeguard import typechecked

from qiboconnection.typings.devices import QuantumDeviceInput

from .online_device import OnlineDevice
from .quantum_device_calibration_details import CalibrationDetails
from .quantum_device_characteristics import QuantumDeviceCharacteristics

# pylint: disable=no-member


class QuantumDevice(OnlineDevice):
    """Quantum Device class"""

    @typechecked
    def __init__(self, device_input: QuantumDeviceInput):
        self._characteristics = (
            QuantumDeviceCharacteristics(device_input.q_characteristics)
            if device_input.q_characteristics is not None
            else None
        )
        self._last_calibration_time = device_input.last_calibration_time
        self._calibration_details = (
            CalibrationDetails(device_input.q_calibration_details)
            if device_input.q_calibration_details is not None
            else None
        )

        super().__init__(device_input)

        self._str = (
            f"<Quantum Device: device_id={self._device_id}, device_name='{self._device_name}'"  # type: ignore[attr-defined]
            f", status='{self._status}'"
        )
        if self._last_calibration_time:
            self._str += f", last_calibration_time={self._last_calibration_time}"
        if self._characteristics:
            self._str += f", characteristics={str(self._characteristics)}"
        if self._calibration_details:
            self._str += f", calibration_details={str(self._calibration_details)}"
        self._str += ">"

    @property
    def __dict__(self):
        """Dictionary representation of a QuantumDevice

        Returns:
            dict: Output dictionary of a QuantumDevice object
        """
        device_dict = super().__dict__

        if self._last_calibration_time:
            device_dict |= {"last_calibration_time": self._last_calibration_time}
        if self._characteristics:
            device_dict |= {"characteristics": self._characteristics.__dict__}
        if self._calibration_details:
            device_dict |= {"calibration_details": self._calibration_details.__dict__}

        return device_dict



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" JobListingItem class """

from dataclasses import field

from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.responses import JobListingItemResponse
from qiboconnection.util import from_kwargs


class JobListingItem:  # pylint: disable=too-few-public-methods
    """JobListing single item representation."""

    user_id: int
    device_id: int
    status: str | JobStatus
    job_type: str | JobType
    number_shots: int
    id: int | None = field(default=None)

    @classmethod
    def from_response(cls, response: JobListingItemResponse):
        """Constructor for JobListingItems that takes in a JobListingItemResponse"""
        return from_kwargs(cls, **response.__dict__)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Quantum Calibration Details """

from typeguard import typechecked

from qiboconnection.typings.devices import CalibrationDetailsInput

from .device_details import DeviceDetails


class CalibrationDetails(DeviceDetails):  # pylint: disable=too-few-public-methods
    """Class representation of Quantum Calibration Details"""

    @typechecked
    def __init__(self, characteristics_input: CalibrationDetailsInput):
        super().__init__()

        self._elapsed_time = characteristics_input.elapsed_time
        self._t1 = characteristics_input.t1
        self._frequency = characteristics_input.frequency

        self._str = "<CalibrationDetails:"
        if self._elapsed_time:
            self._str += f" elapsed_time='{self._elapsed_time}'"
        if self._t1:
            self._str += f" t1='{self._t1}'"
        if self._frequency:
            self._str += f" frequency='{self._frequency}'"
        self._str += ">"

    @property
    def __dict__(self):
        """Dictionary representation of SimulatorDeviceCharacteristics

        Returns:
            dict: Output dictionary of SimulatorDeviceCharacteristics object
        """
        calibration_dict: dict = {}

        if self._elapsed_time:
            calibration_dict |= {"elapsed_time": self._elapsed_time}
        if self._t1:
            calibration_dict |= {"t1": self._t1}
        if self._frequency:
            calibration_dict |= {"frequency": self._frequency}
        return calibration_dict



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Device class """
import json

from requests import HTTPError
from typeguard import typechecked

from qiboconnection.config import logger
from qiboconnection.connection import Connection
from qiboconnection.typings.devices import DeviceInput
from qiboconnection.typings.enums import DeviceAvailability, DeviceStatus

from .device_details import DeviceDetails

# pylint: disable=no-member
# pylint: disable=attribute-defined-outside-init


class Device(DeviceDetails):
    """Abstract class for devices."""

    @typechecked
    def __init__(self, device_input: DeviceInput):
        super().__init__()
        for k, v in vars(device_input).items():
            setattr(self, f"_{k}", v)
        self._str = (
            f"<Device: device_id={self._device_id}, device_name='{self._device_name}', "  # type: ignore[attr-defined]
            + f"status='{self._status}', availability='{self._availability}', channel_id={self._channel_id}>"  # type: ignore[attr-defined]
        )

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        """Returns device identifier

        Returns:
            int: device identifier
        """
        return self._device_id  # type: ignore[attr-defined]

    @property
    def name(self) -> str:
        """Returns device name

        Returns:
            str: device name
        """
        return self._device_name  # type: ignore[attr-defined]

    def block_device(self, connection: Connection) -> None:
        """Blocks a device to avoid others to use it

        Args:
            connection (Connection): Qibo API connection

        Raises:
            HTTPError: Error blocking device
        """
        try:
            connection.update_device_availability(device_id=self._device_id, availability=DeviceAvailability.BLOCKED)  # type: ignore[attr-defined]
            self._availability = DeviceAvailability.BLOCKED
        except HTTPError as ex:
            logger.error("Error blocking device %s.", self._device_name)  # type: ignore[attr-defined]
            raise ex

    def release_device(self, connection: Connection) -> None:
        """Releases a device to let others use it

        Args:
            connection (Connection): Qibo API connection
        """
        connection.update_device_availability(device_id=self._device_id, availability=DeviceAvailability.AVAILABLE)  # type: ignore[attr-defined]
        self._availability = DeviceAvailability.AVAILABLE

    def set_to_online(self, connection: Connection) -> None:
        """Updates a device status so that it can accept remote jobs. Local jobs will be blocked.

        Args:
            connection (Connection): Qibo API connection
        """
        connection.update_device_status(device_id=self._device_id, status=DeviceStatus.ONLINE)  # type: ignore[attr-defined]
        self._status = DeviceStatus.ONLINE

    def set_to_maintenance(self, connection: Connection) -> None:
        """Puts a device in maintenance mode, so that it can only accept local jobs.
         Incoming remote jobs will be blocked.

        Args:
            connection (Connection): Qibo API connection
        """
        connection.update_device_status(device_id=self._device_id, status=DeviceStatus.MAINTENANCE)  # type: ignore[attr-defined]
        self._status = DeviceStatus.MAINTENANCE

    @property
    def __dict__(self):
        """Dictionary representation of a Device

        Returns:
            dict: Output dictionary of a Device object
        """
        return {
            "device_id": self._device_id,  # type: ignore[attr-defined]
            "device_name": self._device_name,  # type: ignore[attr-defined]
            "status": self._status,
            "availability": self._availability,
        }

    def toJSON(self) -> str:  # pylint: disable=invalid-name
        """JSON representation of a Device

        Returns:
            str: JSON serialization of a Device object
        """

        return json.dumps(self.__dict__, indent=2)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Remote Connection """
import json
import os
from abc import ABC
from dataclasses import asdict, dataclass
from datetime import datetime, timezone
from io import TextIOWrapper
from typing import Any, List, Optional, TextIO, Tuple, Union

import jwt
import requests
from typeguard import typechecked

from qiboconnection import __version__ as VERSION  # pylint: disable=cyclic-import
from qiboconnection.config import get_environment, logger
from qiboconnection.errors import ConnectionException, HTTPError, RemoteExecutionException
from qiboconnection.models.user import User
from qiboconnection.typings.connection import ConnectionConfiguration, ConnectionEstablished
from qiboconnection.typings.requests import AssertionPayload
from qiboconnection.typings.responses import AccessTokenResponse
from qiboconnection.util import base64url_encode, process_response


def TIMEOUT():
    """Returns always the last TIMEOUT env var that the user might have set."""
    return int(os.getenv("QIBOCONNECTION_TIMEOUT", "10"))


def refresh_token_if_unauthorised(func):
    """Decorator that, if an HttpError is raised during a call, will retry to perform the call after
    updating the AccessToken.

    Args:
        func: function to decorate
    """

    def decorated(self: "Connection", *args, **kwargs):
        """decorated"""
        try:
            return func(self, *args, **kwargs)
        except HTTPError as ex:
            if ex.response.status_code not in [400, 401]:
                raise ex
            self.update_authorisation_using_refresh_token()
            return func(self, *args, **kwargs)

    return decorated


@dataclass
class Connection(ABC):  # pylint: disable=too-many-instance-attributes
    """Class to create a remote connection to a Qibo server"""

    @typechecked
    def __init__(
        self,
        configuration: ConnectionConfiguration,
        api_path: Optional[str] = None,
    ):
        self._environment = get_environment()
        self._api_path = api_path
        self._remote_server_api_url: str | None = None
        self._remote_server_base_url: str | None = None
        self._authorisation_server_api_call: str | None = None
        self._authorisation_server_refresh_api_call: str | None = None
        self._user_slack_id: Union[str, None] = None
        self._audience_url = f"{self._environment.audience_url}/api/v1"
        self._user: User | None = None
        self._authorisation_access_token: str | None = None
        self._authorisation_refresh_token: str | None = None
        self._load_configuration(configuration, api_path)

    @property
    def user(self) -> User:
        """Gets User

        Returns:
            User: User associated to the connection
        """
        if self._user is None:
            raise ValueError("user not defined")
        return self._user

    @property
    def username(self) -> str:
        """Gets user name

        Returns:
            str: user name associated to the connection
        """
        if self._user is None:
            raise ValueError("user not defined")
        return self._user.username

    @property
    def _user_id(self) -> int | None:
        """Gets user name

        Returns:
            str: user name associated to the connection
        """
        if self._user is None:
            raise ValueError("user not defined")
        return self._user.user_id

    @_user_id.setter
    def _user_id(self, new_id) -> None:
        """Sets user name

        Returns:
            None
        """
        if self._user is None:
            raise ValueError("user not defined")
        self._user.user_id = new_id

    def _load_user_id_from_token(self, access_token):
        self._user_id = jwt.decode(access_token, options={"verify_signature": False})["user_id"]

    @property
    def user_slack_id(self) -> str:
        """Gets user slack id

        Returns:
            str: user slack id associated to the user's connection
        """
        if self._user_slack_id is None:
            self._user_slack_id = self._retrieve_user_slack_id()
        return self._user_slack_id

    def _retrieve_user_slack_id(self) -> str:
        """
        Uses user info to recover the id and make a request to retrieve from server for the User's Slack Id.
        Returns:
        """
        if self._user is None:
            raise ValueError("user not defined")
        user_response, response_status = self.send_get_auth_remote_api_call(path=f"/users/{self._user_id}")
        if response_status != 200:
            raise ValueError(f"Error getting user: {response_status}")
        if "slack_id" not in user_response or user_response["slack_id"] is None:
            return ""
        return user_response["slack_id"]

    def _set_api_calls(self, api_path: str):
        """
        Builds api path, remote server urls and auth server url from env info and api_path kwarg.
        Args:
            api_path: path to api, with leading slash (/).

        Returns:

        """
        self._api_path = api_path
        self._remote_server_api_url = f"{self._environment.qibo_quantum_service_url}{api_path}"
        self._remote_server_base_url = f"{self._environment.qibo_quantum_service_url}"
        self._authorisation_server_api_call = f"{self._remote_server_api_url}/authorisation-tokens"
        self._authorisation_server_refresh_api_call = f"{self._remote_server_api_url}/authorisation-tokens/refresh"

    def _add_version_header(self, header):
        header["X-Client-Version"] = VERSION
        return header

    def _load_configuration(
        self,
        input_configuration: ConnectionConfiguration,
        api_path: Optional[str] = None,
    ) -> None:
        if api_path is None:
            raise ConnectionException("No api path provided.")
        self._set_api_calls(api_path=api_path)
        self._register_configuration_and_request_authorisation_access_token(input_configuration)
        self._load_user_id_from_token(access_token=self._authorisation_access_token)

    def _register_configuration_and_request_authorisation_access_token(self, configuration: ConnectionConfiguration):
        """
        Saves the connection info of user and calls urls, and requests for a new access token to be saved to
         self._authorisation_access_token.
        Args:
            configuration: configuration to save.
        """
        self._register_connection_configuration(configuration)
        self._authorisation_access_token, self._authorisation_refresh_token = self._request_authorisation_token()

    def _register_configuration_with_authorisation_tokens(self, configuration: ConnectionEstablished):
        """
        Saves the connection info of user and calls urls, and saves to self._authorisation_access_token the access
         token.
        Args:
            configuration: configuration to save.
        """
        logger.debug("Configuration loaded successfully.")
        self._register_connection_established(configuration)
        self._authorisation_access_token = configuration.authorisation_access_token
        self._authorisation_refresh_token = configuration.authorisation_refresh_token

    def _register_connection_established(self, configuration: ConnectionEstablished):
        """
        Saves to self._user the provided configuration, and sets the api calls using the api_path inside Configuration.
        Args:
            configuration: configuration to save.
        """
        self._register_connection_configuration(configuration)
        self._set_api_calls(api_path=configuration.api_path)

    def _register_connection_configuration(self, configuration: Union[ConnectionConfiguration, ConnectionEstablished]):
        self._user = User(
            user_id=configuration.user_id,
            username=configuration.username,
            api_key=configuration.api_key,
        )

    def update_device_status(self, device_id: int, status: str) -> Tuple[Any, int]:
        """Update a Device status

        Args:
            device_id (int): device identifier
            status (str): device status

        Returns:
            Tuple[Any, int]: Http Response
        """
        return self.send_put_auth_remote_api_call(path=f"/devices/{device_id}", data={"status": status})

    def update_device_availability(self, device_id: int, availability: str) -> Tuple[Any, int]:
        """Makes the request for updating a Device availability with a new value.

        Args:
            device_id (int): device identifier
            availability (str): device availability

        Returns:
            Tuple[Any, int]: Http Response
        """
        return self.send_put_auth_remote_api_call(path=f"/devices/{device_id}", data={"availability": availability})

    def send_message(self, channel_id: int, message: dict) -> Tuple[Any, int]:
        """Sends a message to a channel registered in the system

        Args:
            channel_id (int): channel identifier
            message (dict): message to send

        Returns:
            Tuple[Any, int]: Http response
        """
        return self.send_post_auth_remote_api_call(path=f"/messages?channel={channel_id}", data=message)

    @refresh_token_if_unauthorised
    @typechecked
    def send_post_auth_remote_api_call(self, path: str, data: Any, timeout: int | None = None) -> Tuple[Any, int]:
        """HTTP POST REST API authenticated call to remote server

        Args:
            path (str): path to add to the remote server api url
            data (Any): data to send
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        response = requests.post(
            f"{self._remote_server_api_url}{path}", json=data.copy(), headers=header, timeout=timeout
        )
        return process_response(response)

    def send_file(
        self, channel_id: int, file: TextIOWrapper, filename: str, timeout: int | None = None
    ) -> Tuple[Any, int]:
        """Sends a file to a channel registered in the system

        Args:
            channel_id (int): channel identifier
            file (TextIOWrapper): file to send
            filename (str): file name
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        return self.send_post_file_auth_remote_api_call(
            path=f"/files?channel={channel_id}", file=file, filename=filename, timeout=timeout
        )

    @refresh_token_if_unauthorised
    @typechecked
    def send_put_auth_remote_api_call(self, path: str, data: Any, timeout: int | None = None) -> Tuple[Any, int]:
        """HTTP PUT REST API authenticated call to remote server

        Args:
            path (str): path to add to the remote server api url
            data (Any): data to send
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        response = requests.put(
            f"{self._remote_server_api_url}{path}", json=data.copy(), headers=header, timeout=timeout
        )
        return process_response(response)

    @refresh_token_if_unauthorised
    @typechecked
    def send_post_file_auth_remote_api_call(
        self, path: str, file: Union[TextIOWrapper, TextIO], filename: str, timeout: int | None = None
    ) -> Tuple[Any, int]:
        """HTTP POST REST API authenticated call to send a file to remote server

        Args:
            path (str): path to add to the remote server api url
            file (Union[TextIOWrapper, TextIO]): file to send
            filename (str): file to send
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        packed_file = {"file": (filename, file)}
        response = requests.post(
            f"{self._remote_server_api_url}{path}", files=packed_file, headers=header, timeout=timeout
        )
        return process_response(response)

    @refresh_token_if_unauthorised
    @typechecked
    def send_get_auth_remote_api_call(
        self, path: str, params: dict | None = None, timeout: int | None = None
    ) -> Tuple[Any, int]:
        """HTTP GET REST API authenticated call to remote server

        Args:
            path (str): path to add to the remote server api url
            params (str): dict of parameters to be encoded as url query params
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        response = requests.get(f"{self._remote_server_api_url}{path}", headers=header, params=params, timeout=timeout)

        if response.status_code != 200:
            error_details = response.json()
            if "detail" in error_details and "does not exist" in error_details["detail"]:
                raise RemoteExecutionException("The job does not exist!", status_code=400)

        return process_response(response)

    @refresh_token_if_unauthorised
    @typechecked
    def send_get_auth_remote_api_call_all_pages(
        self, path: str, params: dict | None = None, timeout: int | None = None
    ) -> List[Tuple[Any, int]]:
        """HTTP GET REST API authenticated call to remote server

        Args:
            path (str): path to add to the remote server api url
            params (str): dict of parameters to be encoded as url query params
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        next_url = f"{self._remote_server_api_url}{path}"
        responses = []
        while "None" not in next_url:
            response = requests.get(next_url, headers=header, params=params, timeout=timeout)
            json_content, status_code = process_response(response)
            next_url = json_content["links"]["next"]
            responses.append((json_content, status_code))
        return responses

    @refresh_token_if_unauthorised
    @typechecked
    def send_delete_auth_remote_api_call(self, path: str, timeout: int | None = None) -> Tuple[Any, int]:
        """HTTP DELETE REST API authenticated call to remote server

        Args:
            path (str): path to add to the remote server api url
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        header = self._add_version_header({"Authorization": f"Bearer {self._authorisation_access_token}"})
        response = requests.delete(f"{self._remote_server_api_url}{path}", headers=header, timeout=timeout)

        if response.status_code != 204:
            error_details = response.json()
            if "detail" in error_details and "does not exist" in error_details["detail"]:
                raise RemoteExecutionException("The job does not exist!", status_code=400)
            response.raise_for_status()

        return ("", 204)

    @refresh_token_if_unauthorised
    @typechecked
    def send_get_remote_call(self, path: str, timeout: int | None = None) -> Tuple[Any, int]:
        """HTTP GET REST API call to remote server (without authentication)

        Args:
            path (str): path to add to the remote server api url
            timeout (int): time to wait. If not provided, a default will be used.

        Returns:
            Tuple[Any, int]: Http response
        """
        timeout = timeout or TIMEOUT()
        logger.debug("Calling: %s%s", self._remote_server_api_url, path)
        response = requests.get(
            f"{self._remote_server_base_url}{path}", timeout=timeout, headers=self._add_version_header({})
        )
        return process_response(response)

    def _request_authorisation_token(self, timeout: int | None = None):
        """
        Builds assertion payload with user info, encodes it and uses it to POST the server for a new Access Token.
        Args:
            timeout (int): time to wait. If not provided, a default will be used.
        Returns: str tuple with new Access  and Refresh Tokens.
        """
        timeout = timeout or TIMEOUT()
        assertion_payload = AssertionPayload(
            **self._user.__dict__,  # type: ignore
            audience=self._audience_url,
            iat=int(datetime.now(timezone.utc).timestamp()),
        )

        encoded_assertion_payload = base64url_encode(json.dumps(asdict(assertion_payload), indent=2))

        authorisation_request_payload = {
            "grantType": "urn:ietf:params:oauth:grant-type:jwt-bearer",
            "assertion": encoded_assertion_payload,
            "scope": "user profile",
        }

        if self._authorisation_server_api_call is None:
            raise ValueError("Authorisation server api call is required")
        logger.debug("Calling: %s", self._authorisation_server_api_call)
        response: requests.Response = requests.post(
            self._authorisation_server_api_call,
            json=authorisation_request_payload,
            timeout=timeout,
            headers=self._add_version_header({}),
        )
        if response.status_code not in [200, 201]:
            try:
                json_content = json.loads(response.content)
                detail = json_content["detail"]
            except (TypeError, KeyError, json.JSONDecodeError):
                detail = ""
            reason_text = f" Reason: {response.reason}." if response.reason else ""
            details_text = f" Details: {detail}." if detail else ""
            raise ValueError(f"Authorisation request failed.{reason_text}{details_text}")

        access_token_response: AccessTokenResponse = AccessTokenResponse(**response.json())
        logger.debug("Connection successfully established.")

        return access_token_response.accessToken, access_token_response.refreshToken

    def update_authorisation_using_refresh_token(self, timeout: int | None = None):
        """Updates the saved access token sending the request token. For this, it
        builds assertion payload with user info, encodes it and uses it to POST the server for a new Access Token.
        Args:
            timeout (int): time to wait. If not provided, a default will be used.
        Returns:
            str with a new Access Token
        """
        timeout = timeout or TIMEOUT()

        if self._authorisation_server_refresh_api_call is None:
            raise ValueError("Authorisation server api call is required")
        logger.debug("Calling: %s", self._authorisation_server_refresh_api_call)
        response: requests.Response = requests.post(
            self._authorisation_server_refresh_api_call,
            json={},
            headers=self._add_version_header({"Authorization": f"Bearer {self._authorisation_refresh_token}"}),
            timeout=timeout,
        )
        if response.status_code not in [200, 201]:
            raise ValueError(f"Authorisation request failed: {response.reason},{response.status_code}")
        logger.debug("Connection successfully renewed.")
        self._authorisation_access_token = AccessTokenResponse(**response.json()).accessToken



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Handling Error Utility Functions """
import json
from typing import Union

from requests.models import HTTPError, Response

from qiboconnection.config import logger


class RemoteExecutionException(Exception):
    """Exception raised when calling remote server

    Args:
        Exception (Exception): Inherit from Exception
    """

    def __init__(self, message: Union[dict, str], status_code: int):
        super().__init__(message)
        self.status_code = status_code
        logger.error("RemoteExecutionException: %s, %i", message, status_code)


class ConnectionException(Exception):
    """Exception raised when establishing the connection to a remote server

    Args:
        Exception (Exception): Inherit from Exception
    """


def custom_raise_for_status(response: Response):
    """Raises :class:`HTTPError`, if one occurred."""

    http_error_msg = ""
    if isinstance(response.reason, bytes):
        # We attempt to decode utf-8 first because some servers
        # choose to localize their reason strings. If the string
        # isn't utf-8, we fall back to iso-8859-1 for all other
        # encodings. (See PR #3538)
        try:
            reason = response.reason.decode("utf-8")
        except UnicodeDecodeError:
            reason = response.reason.decode("iso-8859-1")
    else:
        reason = response.reason

    if 400 <= response.status_code < 500:
        http_error_msg = f"{response.status_code} Client Error: {reason} for url: {response.url}"

    elif 500 <= response.status_code < 600:
        http_error_msg = f"{response.status_code} Server Error: {reason} for url: {response.url}"

    if http_error_msg and response.text:
        try:
            json_text = json.loads(response.text)
            if "detail" in json_text:
                json_text["detail"] += f" {http_error_msg}"
            logger.error(json.dumps(json_text, indent=2))
            raise HTTPError(json.dumps(json_text, indent=2), response=response)
        except Exception as ex:
            json_text_str = str(response.text)
            logger.error(json_text_str)
            raise HTTPError(json_text_str, response=response) from ex

    if http_error_msg:
        raise HTTPError(http_error_msg, response=response)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Quantum Device Typing """
from dataclasses import InitVar, dataclass

from .calibration_details_input import CalibrationDetailsInput
from .online_device_input import OnlineDeviceInput
from .quantum_device_characteristics_input import QuantumDeviceCharacteristicsInput


@dataclass(kw_only=True)
class QuantumDeviceInput(OnlineDeviceInput):
    """Quantum Device Input"""

    last_calibration_time: str | None = ""
    characteristics: InitVar[dict | None] = None
    calibration_details: InitVar[dict | None] = None

    def __post_init__(self, characteristics, calibration_details):
        self._characteristics = QuantumDeviceCharacteristicsInput(**characteristics)
        self._calibration_details = CalibrationDetailsInput(**calibration_details)

    @property
    def q_characteristics(self):
        """Characteristics getter"""
        return self._characteristics

    @property
    def q_calibration_details(self):
        """Calibration details getter"""
        return self._calibration_details



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simulator Device Characteristics"""
import json

from typeguard import typechecked

from qiboconnection.models.devices.device_details import DeviceDetails
from qiboconnection.typings.devices import SimulatorDeviceCharacteristicsInput
from qiboconnection.typings.enums import DeviceType


class SimulatorDeviceCharacteristics(DeviceDetails):
    """Class representation of a Simulator Device Characteristics"""

    @typechecked
    def __init__(self, characteristics_input: SimulatorDeviceCharacteristicsInput):
        super().__init__()
        if (
            characteristics_input.type is not DeviceType.SIMULATOR
            and characteristics_input.type != DeviceType.SIMULATOR
        ):
            raise TypeError("Characteristics Device not supported")

        self._type = characteristics_input.type
        self._cpu = characteristics_input.cpu
        self._gpu = characteristics_input.gpu
        self._os = characteristics_input.os
        self._kernel = characteristics_input.kernel
        self._ram = characteristics_input.ram

        self._str = f"<SimulatorDeviceCharacteristics: type='{self._type}'"
        self._str += f" cpu='{self._cpu}'"
        self._str += f" gpu='{self._gpu}'"
        self._str += f" os='{self._os}'"
        self._str += f" kernel='{self._kernel}'"
        self._str += f" ram='{self._ram}'"
        self._str += ">"

    @property
    def __dict__(self):
        """Dictionary representation of SimulatorDeviceCharacteristics

        Returns:
            dict: Output dictionary of SimulatorDeviceCharacteristics object
        """
        return {
            "type": self._type,
            "cpu": self._cpu,
            "gpu": self._gpu,
            "os": self._os,
            "kernel": self._kernel,
            "ram": self._ram,
        }

    def toJSON(self) -> str:  # pylint: disable=invalid-name
        """JSON representation of SimulatorDeviceCharacteristics

        Returns:
            str: JSON serialization of SimulatorDeviceCharacteristics object
        """

        return json.dumps(self.__dict__, indent=2)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" JobListingItemResponse typing """

from dataclasses import dataclass, field

from qiboconnection.util import from_kwargs


@dataclass
class JobListingItemResponse:
    """Job Response without the results. Includes all jobs metadata so that
    the user can identify the id from the job he is interested to retrieve the results.

    Attributes:
        user_id (int): User identifier
        device_id (int): Device identifier

        id (int): Job identifier
        status (str): Status of the job
    """

    status: str
    user_id: int
    device_id: int
    job_type: str
    number_shots: int
    id: int | None = field(default=None)

    @classmethod
    def from_kwargs(cls, **kwargs):
        """Returns an instance of the class including non-typed attributes"""
        return from_kwargs(cls, **kwargs)

    def to_dict(self):
        """Convert into dict, including non-typed attributes"""
        return self.__dict__



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Online Device """

from typeguard import typechecked

from qiboconnection.typings.devices import OnlineDeviceInput

from .device import Device

# pylint: disable=no-member


class OnlineDevice(Device):
    """Online Device class"""

    @typechecked
    def __init__(self, device_input: OnlineDeviceInput):
        self._number_pending_jobs = None

        super().__init__(device_input)

        self._str = (
            f"<Online Device: device_id={self._device_id},"  # type: ignore[attr-defined]
            f" device_name='{self._device_name}',"  # type: ignore[attr-defined]
            f" status='{self._status}"
        )
        if self._number_pending_jobs is not None:
            self._str += f", number_pending_jobs={self._number_pending_jobs}"
        self._str += ">"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" JobStatus enum """
from .str_enum import StrEnum


class JobStatus(StrEnum):
    """Job Status

    Args:
        enum (str): Accepted values are:
            * "not sent"
            * "pending"
            * "running"
            * "completed"
            * "queued"
            * "error"
    """

    NOT_SENT = "not sent"
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    ERROR = "error"
    QUEUED = "queued"
    CANCELLED = "cancelled"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Utility functions """
import base64
import binascii
import io
import json
import pickle  # nosec - temporary bandit ignore
from base64 import urlsafe_b64decode, urlsafe_b64encode
from inspect import signature
from json.decoder import JSONDecodeError
from typing import Any, List, Tuple

import requests
from qibo.states import CircuitResult

from qiboconnection.errors import custom_raise_for_status


def base64url_encode(payload: dict | bytes | str) -> str:
    """Encode a given payload to base64 string

    Args:
        payload ( dict | bytes | str): data to be encoded

    Returns:
        str: base64 encoded data
    """
    if isinstance(payload, dict):
        payload = json.dumps(payload)
    if not isinstance(payload, bytes):
        payload = payload.encode("utf-8")
    return urlsafe_b64encode(payload).decode("utf-8")


def base64_decode(encoded_data: str) -> str:
    """Decodes a base64 encoded string

    Args:
        encoded_data (str): a base64 encoded string

    Returns:
        Any: The data decoded
    """
    return urlsafe_b64decode(encoded_data).decode("utf-8")


def decode_jsonified_dict(http_response: str) -> dict:
    """Decodes results that have been jsonified and base64 encoded."""
    return json.loads(urlsafe_b64decode(http_response))


def _decode_pickled_results(http_response: str) -> Any:
    """Decodes results that have been pickled and base64 encoded."""
    decoded_result_str = urlsafe_b64decode(http_response)
    result_bytes = io.BytesIO(decoded_result_str)
    return pickle.loads(result_bytes.getbuffer())  # nosec - temporary bandit ignore


def decode_results_from_circuit(http_response: str) -> CircuitResult | dict:
    """Decode the results from the circuit execution. Ideally we should always expect dictionaries here, but for qibo we
    are still serializing `CircuitResult`s that must be pickled.

    Args:
        http_response (str): the execution results as an Http Response

    Returns:
        List[CircuitResult]: a Qibo CircuitResult
    """
    try:
        return decode_jsonified_dict(http_response)
    except (binascii.Error, UnicodeDecodeError, JSONDecodeError):
        return _decode_pickled_results(http_response)


def decode_results_from_qprogram(http_response: str):
    from qililab.result.qprogram_results import QProgramResults

    """Decode the results from the Qililab experiment execution

    Args:
        http_response (str): the execution results as an Http Response

    Returns:
        dict: object containing a serialized representation of a qililab Results object
    """
    return QProgramResults.from_dict(decode_jsonified_dict(http_response)["attributes"])


def process_response(response: requests.Response) -> Tuple[Any, int]:
    """Process an Http Response to check for errors

    Args:
        response (requests.Response): Http Response

    Returns:
        Tuple[Any, int]: Data from the Response, and status code
    """
    custom_raise_for_status(response)
    try:
        return response.json(), response.status_code
    except JSONDecodeError:
        return response.text, response.status_code


def jsonify_dict_and_base64_encode(object_to_encode: dict) -> str:
    """
    Jsonifies a given dict, encodes it to bytes assuming utf-8, and encodes that byte obj to an url-save base64 str
    """
    return str(base64.urlsafe_b64encode(json.dumps(object_to_encode).encode("utf-8")), "utf-8")


def jsonify_list_with_str_and_base64_encode(object_to_encode: List[str]) -> str:
    """Encodes a given list of strings to bytes assuming utf-8, and encodes that byte-array to an url-save base64 str"""
    return str([str(base64.urlsafe_b64encode(s.encode("utf-8")), "utf-8") for s in object_to_encode])


def unzip(zipped_list: List[Tuple[Any, Any]]):
    """Inverse of the python builtin `zip` operation"""
    return tuple(zip(*zipped_list))


def from_kwargs(cls, **kwargs: dict):
    """
    Create an instance of the class by extracting attributes from keyword arguments.

    This method takes keyword arguments and initializes an instance of the class
    with attributes that match the class's constructor parameters. Any additional
    keyword arguments that don't correspond to class attributes are assigned as
    attributes to the created instance.

    Args:
        cls: The class (typically, the class that defines this method).
        **kwargs: Keyword arguments to initialize the instance.

    Returns:
        An instance of the class with attributes initialized from the keyword
        arguments.
    """
    cls_fields = set(signature(cls).parameters)
    native_args, new_args = {}, {}

    for name, val in kwargs.items():
        if name in cls_fields:
            native_args[name] = val
        else:
            new_args[name] = val

    ret = cls(**native_args)

    for new_name, new_val in new_args.items():
        setattr(ret, new_name, new_val)
    return ret



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Algorithm Typing """

from dataclasses import dataclass
from typing import List, Optional

from qiboconnection.typings.enums import AlgorithmName, AlgorithmType, InitialValue


@dataclass
class AlgorithmOptions:
    """Algorithm Options data class"""

    number_qubits: int
    initial_value: Optional[InitialValue] = InitialValue.ZERO

    @property
    def __dict__(self):
        """
        Build class dict.
        Returns:
            dict: representation of the class
        """
        return {"number_qubits": self.number_qubits, "initial_value": self.initial_value.value}


@dataclass
class AlgorithmDefinition:
    """Algorithm data class"""

    name: AlgorithmName
    algorithm_type: AlgorithmType
    options: AlgorithmOptions

    @property
    def __dict__(self):
        """
        Build class dict.
        Returns:
            dict: representation of the class
        """
        return {
            "name": self.name.value,
            "type": self.algorithm_type.value,
            "options": self.options.__dict__,
        }


@dataclass
class ProgramDefinition:
    """Program data class"""

    algorithms: List[AlgorithmDefinition]



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Util Functions used by the API module """

import ast
import json
from typing import Any, List

from qibo.models.circuit import Circuit

from qiboconnection.config import logger
from qiboconnection.models import JobResult
from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.util import base64_decode


def parse_job_responses_to_results(job_responses: List[JobResponse]) -> List[dict | Any | None]:
    """Parse a list of job_responses to a list of dict with the content of each job. If the job is not COMPLETED,
    put a None in its place. For this, we build a JobResult instance for each COMPLETED job, and then we keep its
    `.data`.

    Args:
        job_responses: list of JobResponse instances from which we'll

    Returns:

    """
    raw_results = [
        JobResult(job_id=job_response.job_id, job_type=job_response.job_type, http_response=job_response.result).data
        if job_response.status == JobStatus.COMPLETED
        else None
        for job_response in job_responses
    ]
    return list(raw_results)


def deserialize_job_description(base64_description: str, job_type: str) -> list[Circuit] | Circuit | dict | str:
    """Convert base64 job description to its corresponding Qibo Circuit or Qililab experiment

    Args:
        base64_description (str):
        job_type (str):

    Raises:
        ValueError: Job type isn't nor Qibo Circuit neither Qililab experiment

    Returns:
        Circuit | dict: _description_
    """
    if job_type == JobType.CIRCUIT:
        try:
            circuits_descriptions = ast.literal_eval(base64_description)
            return [Circuit.from_qasm(base64_decode(encoded_data=description)) for description in circuits_descriptions]

        except SyntaxError:
            return Circuit.from_qasm(base64_decode(encoded_data=base64_description))

    if job_type == JobType.QPROGRAM:
        return json.loads(base64_decode(encoded_data=base64_description))

    logger.warning(f"JobType {job_type} not supported in this Qiboconnection version!")
    return base64_description


def log_job_status_info(job_response: JobResponse):
    """Logs a message depending on the `JobResponse.status` belonging to a certain job.

    Args:
        job_response:

    Returns:

    """
    if job_response.status in [JobStatus.QUEUED, JobStatus.PENDING]:
        logger.warning(
            "Your job with id %i is still pending. Job queue position: %s",
            job_response.job_id,
            job_response.queue_position,
        )
        return None

    if job_response.status == JobStatus.RUNNING:
        logger.warning("Your job with id %i is still running.", job_response.job_id)
        return None
    if job_response.status == JobStatus.NOT_SENT:
        logger.warning("Your job with id %i has not been sent.", job_response.job_id)
        return None
    if job_response.status == JobStatus.ERROR:
        logger.error("Your job with id %i failed.", job_response.job_id)
        return None
    if job_response.status == JobStatus.COMPLETED:
        logger.warning("Your job with id %i is completed.", job_response.job_id)
        return None

    logger.warning(f"Your job with id %i is {job_response.status}.", job_response.job_id)
    return None



# Installation script for python
import os
import re

from setuptools import find_packages, setup

PACKAGE = "qiboconnection"


# Returns the library version
def get_version():
    """Gets the version from the package's __init__ file
    if there is some problem, let it happily fail"""
    VERSIONFILE = os.path.join("src", PACKAGE, "__init__.py")
    initfile_lines = open(VERSIONFILE, "rt").readlines()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    for line in initfile_lines:
        mo = re.search(VSRE, line, re.M)
        if mo:
            return mo.group(1)


# Read in requirements
requirements = open("requirements.txt").readlines()
requirements = [r.strip() for r in requirements]


# load long description from README
this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="qiboconnection",
    version=get_version(),
    description="Python interface to Qilimanjaro's Services for quantum job executions.",
    author="Qilimanjaro",
    author_email="info@qilimanjaro.tech",
    url="https://github.com/qilimanjaro-tech/qiboconnection",
    packages=find_packages("src"),
    package_dir={"": "src"},
    package_data={"": ["*.out"]},
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    install_requires=requirements,
    extras_require={
        "docs": [
            "sphinx",
            "sphinx_rtd_theme",
            "recommonmark",
            "sphinxcontrib-bibtex",
            "sphinx_markdown_tables",
            "nbsphinx",
            "IPython",
        ],
        "tests": ["pytest"],
    },
    python_requires=">=3.10.0",
    long_description=long_description,
    long_description_content_type="text/markdown",
)
